@@ -1,27 +1,24 @@
 //! Encoders and decoders for reading/writing byte sequences.
-use std::io::{Read, Write};
+use std::cmp;
 use std::mem;
 use trackable::error::ErrorKindExt;
 
-use {Decode, DecodeBuf, Encode, EncodeBuf, Error, ErrorKind, ExactBytesEncode, Result};
+use {ByteCount, Decode, Encode, Eos, ErrorKind, ExactBytesEncode, Result};
 
 /// `BytesEncoder` writes the given bytes into an output byte sequence.
 ///
 /// # Examples
 ///
 /// ```
-/// use bytecodec::{Encode, EncodeBuf, EncodeExt};
+/// use bytecodec::{Encode, EncodeExt};
 /// use bytecodec::bytes::BytesEncoder;
+/// use bytecodec::io::IoEncodeExt;
 ///
-/// let mut output = [0; 4];
-/// {
-///     let mut buf = EncodeBuf::new(&mut output);
-///     let mut encoder = BytesEncoder::with_item(b"foo").unwrap();
-///     encoder.encode(&mut buf).unwrap();
-///     assert!(encoder.is_idle());
-///     assert_eq!(buf.len(), 1);
-/// }
-/// assert_eq!(&output[..], b"foo\x00");
+/// let mut output = Vec::new();
+/// let mut encoder = BytesEncoder::with_item(b"foo").unwrap();
+/// encoder.encode_all(&mut output).unwrap();
+/// assert!(encoder.is_idle());
+/// assert_eq!(output, b"foo");
 /// ```
 #[derive(Debug)]
 pub struct BytesEncoder<B> {
@@ -45,15 +42,17 @@ impl<B> Default for BytesEncoder<B> {
 impl<B: AsRef<[u8]>> Encode for BytesEncoder<B> {
     type Item = B;
 
-    fn encode(&mut self, buf: &mut EncodeBuf) -> Result<()> {
+    fn encode(&mut self, buf: &mut [u8], eos: Eos) -> Result<usize> {
+        let mut size = 0;
         let drop_item = if let Some(ref b) = self.bytes {
-            let size = track!(buf.write(&b.as_ref()[self.offset..]).map_err(Error::from))?;
+            size = cmp::min(buf.len(), b.as_ref().len() - self.offset);
+            (&mut buf[..size]).copy_from_slice(&b.as_ref()[self.offset..][..size]);
             self.offset += size;
             if self.offset == b.as_ref().len() {
                 true
             } else {
-                track_assert!(!buf.is_eos(), ErrorKind::UnexpectedEos;
-                              self.offset, b.as_ref().len());
+                track_assert!(!eos.is_reached(), ErrorKind::UnexpectedEos;
+                              buf.len(), size, self.offset, b.as_ref().len());
                 false
             }
         } else {
@@ -62,7 +61,7 @@ impl<B: AsRef<[u8]>> Encode for BytesEncoder<B> {
         if drop_item {
             self.bytes = None;
         }
-        Ok(())
+        Ok(size)
     }
 
     fn start_encoding(&mut self, item: Self::Item) -> Result<()> {
@@ -72,8 +71,8 @@ impl<B: AsRef<[u8]>> Encode for BytesEncoder<B> {
         Ok(())
     }
 
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        Some(self.requiring_bytes())
+    fn requiring_bytes(&self) -> ByteCount {
+        ByteCount::Finite(self.exact_requiring_bytes())
     }
 
     fn is_idle(&self) -> bool {
@@ -81,7 +80,7 @@ impl<B: AsRef<[u8]>> Encode for BytesEncoder<B> {
     }
 }
 impl<B: AsRef<[u8]>> ExactBytesEncode for BytesEncoder<B> {
-    fn requiring_bytes(&self) -> u64 {
+    fn exact_requiring_bytes(&self) -> u64 {
         self.bytes
             .as_ref()
             .map_or(0, |b| b.as_ref().len() - self.offset) as u64
@@ -95,25 +94,24 @@ impl<B: AsRef<[u8]>> ExactBytesEncode for BytesEncoder<B> {
 /// # Examples
 ///
 /// ```
-/// use bytecodec::{Decode, DecodeBuf};
+/// use bytecodec::{Decode, Eos};
 /// use bytecodec::bytes::CopyableBytesDecoder;
 ///
 /// let mut decoder = CopyableBytesDecoder::new([0; 3]);
-/// let mut input = DecodeBuf::new(b"fooba");
+/// let mut input = b"foobar";
 ///
 /// // Decodes first item
-/// assert_eq!(decoder.requiring_bytes_hint(), Some(3));
-/// let item = decoder.decode(&mut input).unwrap();
+/// assert_eq!(decoder.requiring_bytes().to_u64(), Some(3));
+/// let (_, item) = decoder.decode(&input[0..3], Eos::new(false)).unwrap();
 /// assert_eq!(item.as_ref(), Some(b"foo"));
 ///
 /// // Decodes second item
-/// assert_eq!(decoder.requiring_bytes_hint(), Some(3));
-/// let item = decoder.decode(&mut input).unwrap();
+/// assert_eq!(decoder.requiring_bytes().to_u64(), Some(3));
+/// let (_, item) = decoder.decode(&input[3..5], Eos::new(false)).unwrap();
 /// assert_eq!(item, None);
-/// assert_eq!(decoder.requiring_bytes_hint(), Some(1));
+/// assert_eq!(decoder.requiring_bytes().to_u64(), Some(1));
 ///
-/// let mut input = DecodeBuf::new(b"r");
-/// let item = decoder.decode(&mut input).unwrap();
+/// let (_, item) = decoder.decode(&input[5..], Eos::new(true)).unwrap();
 /// assert_eq!(item.as_ref(), Some(b"bar"));
 /// ```
 #[derive(Debug, Default)]
@@ -130,19 +128,19 @@ impl<B> CopyableBytesDecoder<B> {
 impl<B: AsRef<[u8]> + AsMut<[u8]> + Copy> Decode for CopyableBytesDecoder<B> {
     type Item = B;
 
-    fn decode(&mut self, buf: &mut DecodeBuf) -> Result<Option<Self::Item>> {
-        let size = track!(
-            buf.read(&mut self.bytes.as_mut()[self.offset..])
-                .map_err(Error::from)
-        )?;
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+        let size = cmp::min(buf.len(), self.bytes.as_ref().len() - self.offset);
+        (&mut self.bytes.as_mut()[self.offset..][..size]).copy_from_slice(&buf[..size]);
         self.offset += size;
 
         if self.offset == self.bytes.as_mut().len() {
             self.offset = 0;
-            Ok(Some(self.bytes))
+            Ok((size, Some(self.bytes)))
+        } else if eos.is_reached() {
+            track_assert_eq!(self.offset, 0, ErrorKind::UnexpectedEos);
+            Ok((size, None))
         } else {
-            track_assert!(!buf.is_eos(), ErrorKind::UnexpectedEos);
-            Ok(None)
+            Ok((size, None))
         }
     }
 
@@ -150,12 +148,8 @@ impl<B: AsRef<[u8]> + AsMut<[u8]> + Copy> Decode for CopyableBytesDecoder<B> {
         false
     }
 
-    fn is_idle(&self) -> bool {
-        self.offset == 0
-    }
-
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        Some((self.bytes.as_ref().len() - self.offset) as u64)
+    fn requiring_bytes(&self) -> ByteCount {
+        ByteCount::Finite((self.bytes.as_ref().len() - self.offset) as u64)
     }
 }
 
@@ -166,23 +160,23 @@ impl<B: AsRef<[u8]> + AsMut<[u8]> + Copy> Decode for CopyableBytesDecoder<B> {
 /// # Examples
 ///
 /// ```
-/// use bytecodec::{Decode, DecodeBuf};
+/// use bytecodec::Decode;
 /// use bytecodec::bytes::BytesDecoder;
+/// use bytecodec::io::IoDecodeExt;
 ///
 /// let mut decoder = BytesDecoder::new([0; 3]);
-/// assert_eq!(decoder.requiring_bytes_hint(), Some(3));
+/// assert_eq!(decoder.requiring_bytes().to_u64(), Some(3));
 ///
-/// let mut input = DecodeBuf::new(b"foobar");
-/// let item = decoder.decode(&mut input).unwrap();
-/// assert_eq!(item.as_ref(), Some(b"foo"));
-/// assert_eq!(decoder.requiring_bytes_hint(), Some(0)); // no more items are decoded
+/// let item = decoder.decode_exact(b"foobar".as_ref()).unwrap();
+/// assert_eq!(item.as_ref(), b"foo");
+/// assert_eq!(decoder.requiring_bytes().to_u64(), Some(0)); // no more items are decoded
 /// ```
 #[derive(Debug)]
 pub struct BytesDecoder<B> {
     bytes: Option<B>,
     offset: usize,
 }
-impl<B> BytesDecoder<B> {
+impl<B: AsRef<[u8]> + AsMut<[u8]>> BytesDecoder<B> {
     /// Makes a new `BytesDecoder` instance for filling the given byte slice.
     pub fn new(bytes: B) -> Self {
         BytesDecoder {
@@ -190,26 +184,29 @@ impl<B> BytesDecoder<B> {
             offset: 0,
         }
     }
+
+    fn exact_requiring_bytes(&self) -> u64 {
+        self.bytes
+            .as_ref()
+            .map_or(0, |b| b.as_ref().len() - self.offset) as u64
+    }
 }
 impl<B: AsRef<[u8]> + AsMut<[u8]>> Decode for BytesDecoder<B> {
     type Item = B;
 
-    fn decode(&mut self, buf: &mut DecodeBuf) -> Result<Option<Self::Item>> {
-        if let Some(ref mut bytes) = self.bytes {
-            let size = track!(
-                buf.read(&mut bytes.as_mut()[self.offset..])
-                    .map_err(Error::from)
-            )?;
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+        let size = {
+            let mut bytes = track_assert_some!(self.bytes.as_mut(), ErrorKind::DecoderTerminated);
+            let size = cmp::min(buf.len(), bytes.as_ref().len() - self.offset);
+            (&mut bytes.as_mut()[self.offset..][..size]).copy_from_slice(&buf[..size]);
             self.offset += size;
+            size
+        };
+        if self.exact_requiring_bytes() == 0 {
+            Ok((size, self.bytes.take()))
         } else {
-            track_panic!(ErrorKind::DecoderTerminated);
-        }
-
-        if Some(self.offset) == self.bytes.as_ref().map(|b| b.as_ref().len()) {
-            Ok(self.bytes.take())
-        } else {
-            track_assert!(!buf.is_eos(), ErrorKind::UnexpectedEos);
-            Ok(None)
+            track_assert!(!eos.is_reached(), ErrorKind::UnexpectedEos);
+            Ok((size, None))
         }
     }
 
@@ -217,15 +214,8 @@ impl<B: AsRef<[u8]> + AsMut<[u8]>> Decode for BytesDecoder<B> {
         self.bytes.is_none()
     }
 
-    fn is_idle(&self) -> bool {
-        self.offset == 0 || self.bytes.is_none()
-    }
-
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        let n = self.bytes
-            .as_ref()
-            .map_or(0, |b| b.as_ref().len() - self.offset);
-        Some(n as u64)
+    fn requiring_bytes(&self) -> ByteCount {
+        ByteCount::Finite(self.exact_requiring_bytes())
     }
 }
 
@@ -234,23 +224,19 @@ impl<B: AsRef<[u8]> + AsMut<[u8]>> Decode for BytesDecoder<B> {
 /// # Examples
 ///
 /// ```
-/// use bytecodec::{Decode, DecodeBuf};
+/// use bytecodec::{Decode, Eos};
 /// use bytecodec::bytes::RemainingBytesDecoder;
 ///
 /// let mut decoder = RemainingBytesDecoder::new();
-/// assert_eq!(decoder.requiring_bytes_hint(), None);
+/// assert_eq!(decoder.requiring_bytes().to_u64(), None);
 ///
-/// let mut input = DecodeBuf::new(b"foo");
-/// let item = decoder.decode(&mut input).unwrap();
+/// let (size, item) = decoder.decode(b"foo", Eos::new(false)).unwrap();
 /// assert_eq!(item, None);
-/// assert!(input.is_empty());
-/// assert!(!input.is_eos());
+/// assert_eq!(size, 3);
 ///
-/// let mut input = DecodeBuf::with_remaining_bytes(b"bar", 0);
-/// let item = decoder.decode(&mut input).unwrap();
+/// let (size, item) = decoder.decode(b"bar", Eos::new(true)).unwrap();
 /// assert_eq!(item, Some(b"foobar".to_vec()));
-/// assert!(input.is_empty());
-/// assert!(input.is_eos());
+/// assert_eq!(size, 3);
 /// ```
 #[derive(Debug, Default)]
 pub struct RemainingBytesDecoder(Vec<u8>);
@@ -263,16 +249,16 @@ impl RemainingBytesDecoder {
 impl Decode for RemainingBytesDecoder {
     type Item = Vec<u8>;
 
-    fn decode(&mut self, buf: &mut DecodeBuf) -> Result<Option<Self::Item>> {
-        if let Some(additional) = buf.remaining_bytes() {
-            self.0.reserve_exact(buf.len() + additional as usize);
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+        if let Some(remaining) = eos.remaining_bytes().to_u64() {
+            self.0.reserve_exact(buf.len() + remaining as usize);
         }
 
-        track!(buf.read_to_end(&mut self.0).map_err(Error::from))?;
-        if buf.is_eos() {
-            Ok(Some(mem::replace(&mut self.0, Vec::new())))
+        self.0.extend_from_slice(buf);
+        if eos.is_reached() {
+            Ok((buf.len(), Some(mem::replace(&mut self.0, Vec::new()))))
         } else {
-            Ok(None)
+            Ok((buf.len(), None))
         }
     }
 
@@ -280,12 +266,8 @@ impl Decode for RemainingBytesDecoder {
         false
     }
 
-    fn is_idle(&self) -> bool {
-        self.0.is_empty()
-    }
-
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        None
+    fn requiring_bytes(&self) -> ByteCount {
+        ByteCount::Infinite
     }
 }
 
@@ -302,18 +284,15 @@ impl<T: AsRef<str>> AsRef<[u8]> for Utf8Bytes<T> {
 /// # Examples
 ///
 /// ```
-/// use bytecodec::{Encode, EncodeBuf, EncodeExt};
+/// use bytecodec::{Encode, EncodeExt};
 /// use bytecodec::bytes::Utf8Encoder;
+/// use bytecodec::io::IoEncodeExt;
 ///
-/// let mut output = [0; 4];
-/// {
-///     let mut buf = EncodeBuf::new(&mut output);
-///     let mut encoder = Utf8Encoder::with_item("foo").unwrap();
-///     encoder.encode(&mut buf).unwrap();
-///     assert!(encoder.is_idle());
-///     assert_eq!(buf.len(), 1);
-/// }
-/// assert_eq!(&output[..], b"foo\x00");
+/// let mut output = Vec::new();
+/// let mut encoder = Utf8Encoder::with_item("foo").unwrap();
+/// encoder.encode_all(&mut output).unwrap();
+/// assert!(encoder.is_idle());
+/// assert_eq!(output, b"foo");
 /// ```
 #[derive(Debug)]
 pub struct Utf8Encoder<S = String>(BytesEncoder<Utf8Bytes<S>>);
@@ -326,16 +305,16 @@ impl<S> Utf8Encoder<S> {
 impl<S: AsRef<str>> Encode for Utf8Encoder<S> {
     type Item = S;
 
-    fn encode(&mut self, buf: &mut EncodeBuf) -> Result<()> {
-        track!(self.0.encode(buf))
+    fn encode(&mut self, buf: &mut [u8], eos: Eos) -> Result<usize> {
+        track!(self.0.encode(buf, eos))
     }
 
     fn start_encoding(&mut self, item: Self::Item) -> Result<()> {
         track!(self.0.start_encoding(Utf8Bytes(item)))
     }
 
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        self.0.requiring_bytes_hint()
+    fn requiring_bytes(&self) -> ByteCount {
+        self.0.requiring_bytes()
     }
 
     fn is_idle(&self) -> bool {
@@ -343,8 +322,8 @@ impl<S: AsRef<str>> Encode for Utf8Encoder<S> {
     }
 }
 impl<S: AsRef<str>> ExactBytesEncode for Utf8Encoder<S> {
-    fn requiring_bytes(&self) -> u64 {
-        self.0.requiring_bytes()
+    fn exact_requiring_bytes(&self) -> u64 {
+        self.0.exact_requiring_bytes()
     }
 }
 impl<S> Default for Utf8Encoder<S> {
@@ -354,16 +333,16 @@ impl<S> Default for Utf8Encoder<S> {
 }
 
 /// `Utf8Decoder` decodes Rust strings from a input byte sequence.
+///
 /// # Examples
 ///
 /// ```
-/// use bytecodec::{Decode, DecodeBuf};
+/// use bytecodec::{Decode, Eos};
 /// use bytecodec::bytes::Utf8Decoder;
 ///
 /// let mut decoder = Utf8Decoder::new();
 ///
-/// let mut input = DecodeBuf::with_remaining_bytes(b"foo", 0);
-/// let item = decoder.decode(&mut input).unwrap();
+/// let (_, item) = decoder.decode(b"foo", Eos::new(true)).unwrap();
 /// assert_eq!(item, Some("foo".to_owned()));
 /// ```
 #[derive(Debug, Default)]
@@ -389,12 +368,14 @@ where
 {
     type Item = String;
 
-    fn decode(&mut self, buf: &mut DecodeBuf) -> Result<Option<Self::Item>> {
-        if let Some(bytes) = track!(self.0.decode(buf))? {
-            let s = track!(String::from_utf8(bytes).map_err(|e| ErrorKind::InvalidInput.cause(e)))?;
-            Ok(Some(s))
-        } else {
-            Ok(None)
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+        let (size, item) = track!(self.0.decode(buf, eos))?;
+        match item {
+            Some(b) => {
+                let s = track!(String::from_utf8(b).map_err(|e| ErrorKind::InvalidInput.cause(e)))?;
+                Ok((size, Some(s)))
+            }
+            None => Ok((size, None)),
         }
     }
 
@@ -402,46 +383,39 @@ where
         self.0.has_terminated()
     }
 
-    fn is_idle(&self) -> bool {
-        self.0.is_idle()
-    }
-
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        self.0.requiring_bytes_hint()
+    fn requiring_bytes(&self) -> ByteCount {
+        self.0.requiring_bytes()
     }
 }
 
 #[cfg(test)]
 mod test {
-    use {Decode, DecodeBuf, Encode, EncodeBuf, EncodeExt, ErrorKind};
+    use {Encode, EncodeExt, ErrorKind};
+    use io::{IoDecodeExt, IoEncodeExt};
     use super::*;
 
     #[test]
     fn bytes_decoder_works() {
         let mut decoder = BytesDecoder::new([0; 3]);
-        assert_eq!(decoder.requiring_bytes_hint(), Some(3));
+        assert_eq!(decoder.requiring_bytes().to_u64(), Some(3));
 
-        let mut input = DecodeBuf::new(b"foobar");
-        let item = decoder.decode(&mut input).unwrap();
-        assert_eq!(item.as_ref(), Some(b"foo"));
-        assert_eq!(decoder.requiring_bytes_hint(), Some(0));
+        let mut input = b"foobar".as_ref();
+        let item = decoder.decode_exact(&mut input).unwrap();
+        assert_eq!(item.as_ref(), b"foo");
+        assert_eq!(decoder.requiring_bytes().to_u64(), Some(0));
 
         assert_eq!(
-            decoder.decode(&mut input).err().map(|e| *e.kind()),
+            decoder.decode_exact(&mut input).err().map(|e| *e.kind()),
             Some(ErrorKind::DecoderTerminated)
         );
     }
 
     #[test]
     fn utf8_encoder_works() {
-        let mut output = [0; 4];
-        {
-            let mut buf = EncodeBuf::new(&mut output);
-            let mut encoder = Utf8Encoder::with_item("foo").unwrap();
-            encoder.encode(&mut buf).unwrap();
-            assert!(encoder.is_idle());
-            assert_eq!(buf.len(), 1);
-        }
-        assert_eq!(&output[..], b"foo\x00");
+        let mut buf = Vec::new();
+        let mut encoder = Utf8Encoder::with_item("foo").unwrap();
+        encoder.encode_all(&mut buf).unwrap();
+        assert!(encoder.is_idle());
+        assert_eq!(buf, b"foo");
     }
 }
