@@ -1,8 +1,8 @@
 use std;
 
-use {DecodeBuf, Error, ErrorKind, Result};
+use {ByteCount, Eos, Error, Result};
 use combinator::{AndThen, Assert, Collect, DecoderChain, Length, Map, MapErr, MaxBytes, Omit,
-                 SkipRemaining, Take, TryMap};
+                 SkipRemaining, Slice, Take, TryMap};
 
 /// This trait allows for decoding items from a byte sequence incrementally.
 pub trait Decode {
@@ -11,119 +11,82 @@ pub trait Decode {
 
     /// Consumes the given buffer (a part of a byte sequence), and decodes an item from it.
     ///
-    /// If an item is successfully decoded, the decoder will return `Ok(Some(..))`.
+    /// The first element of a succeeded result is the number of bytes consumed
+    /// from `buf` by the decoding process.
+    ///
+    /// If an item is successfully decoded, the decoder will return `Ok((_, Some(..)))`.
     ///
     /// If the buffer does not contain enough bytes to decode the next item,
-    /// the decoder will return `Ok(None)`.
-    /// In this case, the decoder **must** consume all the bytes in the buffer.
+    /// the decoder will return `Ok((_, None))`.
+    /// In this case, the decoder must consume as many bytes in the buffer as possible.
+    ///
+    /// If an item is not yet decoded but the number of consumed bytes in the last `decode` invocation
+    /// is smaller than the length of `buf`, it means the decoder has been suspended its work in any reasons.
+    /// In that case the decoder may require some instructions from clients to resume the work,
+    /// but its concrete method is beyond the scope of this trait.
     ///
     /// # Errors
     ///
     /// Decoders return the following kinds of errors as necessary:
     /// - `ErrorKind::DecoderTerminated`:
     ///   - If all decodable items have been decoded,
-    ///     the decoder **must** return this kind of error when `decode()` method is called.
+    ///     the decoder must return this kind of error when `decode` method is called.
     /// - `ErrorKind::UnexpectedEos`:
-    ///   - `DecodeBuf::is_eos()` returns `true` despite of
+    ///   - The invocation of `eos.is_reached()` returns `true` despite of
     ///     the decoder requires more bytes to decode the next item.
     /// - `ErrorKind::InvalidInput`:
     ///   - Decoded items have invalid values
     ///   - Invalid parameters were given to decoders
     /// - `ErrorKind::Error`:
     ///   - Other errors
-    fn decode(&mut self, buf: &mut DecodeBuf) -> Result<Option<Self::Item>>;
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)>;
 
     /// Returns `true` if the decoder cannot decode items anymore, otherwise `false`.
     ///
     /// If it returns `true`, the next invocation of `decode` method
-    /// **must** return an `ErrorKind::DecoderTerminated` error.
+    /// must return an `ErrorKind::DecoderTerminated` error, and vice versa.
     fn has_terminated(&self) -> bool;
 
-    /// Returns `true` if the decoder does not have an item that being decoded, otherwise `false`.
-    fn is_idle(&self) -> bool;
-
     /// Returns the lower bound of the number of bytes needed to decode the next item.
     ///
-    /// If the decoder does not know the value, it will return `None`
+    /// If the decoder does not know the value, it will return `ByteCount::Unknown`
     /// (e.g., null-terminated strings have no pre-estimable length).
     ///
-    /// If the decoder returns `Some(0)`, it means one of the followings:
+    /// If the decoder returns `ByteCount::Finite(0)`, it means one of the followings:
     /// - (a) There is an already decoded item
     ///   - The next invocation of `decode()` will return it without consuming any bytes
     /// - (b) There are no decodable items
     ///   - All decodable items have been decoded, and the decoder cannot do any further works
-    ///
-    /// The default implementation returns `Some(0)` if the decoder has terminated, otherwise `None`.
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        if self.has_terminated() {
-            Some(0)
-        } else {
-            None
-        }
-    }
+    fn requiring_bytes(&self) -> ByteCount;
 }
-impl<D: ?Sized + Decode> Decode for Box<D> {
+impl<'a, D: ?Sized + Decode> Decode for &'a mut D {
     type Item = D::Item;
 
-    fn decode(&mut self, buf: &mut DecodeBuf) -> Result<Option<Self::Item>> {
-        (**self).decode(buf)
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+        (**self).decode(buf, eos)
     }
 
     fn has_terminated(&self) -> bool {
         (**self).has_terminated()
     }
 
-    fn is_idle(&self) -> bool {
-        (**self).is_idle()
-    }
-
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        (**self).requiring_bytes_hint()
+    fn requiring_bytes(&self) -> ByteCount {
+        (**self).requiring_bytes()
     }
 }
+impl<D: ?Sized + Decode> Decode for Box<D> {
+    type Item = D::Item;
 
-/// `DecodedValue` represents a decoded item.
-///
-/// It does not consume any bytes, and returns the given item when `decode()` was first called.
-///
-/// # Examples
-///
-/// ```
-/// use bytecodec::{Decode, DecodeBuf, DecodedValue};
-///
-/// let mut decoder = DecodedValue::new(10);
-///
-/// let mut input = DecodeBuf::new(b"foo");
-/// let item = decoder.decode(&mut input).unwrap();
-/// assert_eq!(item, Some(10));
-/// assert_eq!(input.len(), 3);
-/// ```
-#[derive(Debug)]
-pub struct DecodedValue<T>(Option<T>);
-impl<T> DecodedValue<T> {
-    /// Makes a new `DecodedValue` instance.
-    pub fn new(value: T) -> Self {
-        DecodedValue(Some(value))
-    }
-}
-impl<T> Decode for DecodedValue<T> {
-    type Item = T;
-
-    fn decode(&mut self, _buf: &mut DecodeBuf) -> Result<Option<Self::Item>> {
-        let item = track_assert_some!(self.0.take(), ErrorKind::DecoderTerminated);
-        Ok(Some(item))
+    fn decode(&mut self, buf: &[u8], eos: Eos) -> Result<(usize, Option<Self::Item>)> {
+        (**self).decode(buf, eos)
     }
 
     fn has_terminated(&self) -> bool {
-        self.0.is_none()
+        (**self).has_terminated()
     }
 
-    fn is_idle(&self) -> bool {
-        self.0.is_none()
-    }
-
-    fn requiring_bytes_hint(&self) -> Option<u64> {
-        Some(0)
+    fn requiring_bytes(&self) -> ByteCount {
+        (**self).requiring_bytes()
     }
 }
 
@@ -134,12 +97,13 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt};
+    /// use bytecodec::{Decode, DecodeExt};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = U8Decoder::new().map(|b| b * 2);
-    /// let item = decoder.decode(&mut DecodeBuf::new(&[10][..])).unwrap();
-    /// assert_eq!(item, Some(20));
+    /// let item = decoder.decode_exact([10].as_ref()).unwrap();
+    /// assert_eq!(item, 20);
     /// ```
     fn map<T, F>(self, f: F) -> Map<Self, T, F>
     where
@@ -157,21 +121,21 @@ pub trait DecodeExt: Decode + Sized {
     /// #[macro_use]
     /// extern crate trackable;
     ///
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt, ErrorKind, Result};
+    /// use bytecodec::{Decode, DecodeExt, ErrorKind, Result};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// # fn main() {
     /// let mut decoder = U8Decoder::new().try_map(|b| -> Result<_> {
     ///     track_assert_ne!(b, 0, ErrorKind::InvalidInput);
     ///     Ok(b * 2)
     /// });
-    /// let mut input = DecodeBuf::new(&[0, 4][..]);
     ///
-    /// let error = decoder.decode(&mut input).err().unwrap();
+    /// let error = decoder.decode_exact([0].as_ref()).err().unwrap();
     /// assert_eq!(*error.kind(), ErrorKind::InvalidInput);
     ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some(8));
+    /// let item = decoder.decode_exact([4].as_ref()).unwrap();
+    /// assert_eq!(item, 8);
     /// # }
     /// ```
     fn try_map<F, T, E>(self, f: F) -> TryMap<Self, F, T, E>
@@ -193,26 +157,27 @@ pub trait DecodeExt: Decode + Sized {
     /// #[macro_use]
     /// extern crate trackable;
     ///
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt};
+    /// use bytecodec::{Decode, DecodeExt};
     /// use bytecodec::fixnum::U16beDecoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// # fn main() {
     /// let mut decoder =
     ///     U16beDecoder::new().map_err(|e| track!(e, "oops!"));
     ///     // or `track_err!(U16beDecoder::new(), "oops!")`
     ///
-    /// let mut input =
-    ///     DecodeBuf::with_remaining_bytes(&[10][..], 0); // Insufficient bytes
-    ///
-    /// let error = track!(decoder.decode(&mut input)).err().unwrap();
+    /// let input = [0]; // Insufficient bytes
+    /// let error = track!(decoder.decode_exact(input.as_ref())).err().unwrap();
     ///
     /// assert_eq!(error.to_string(), "\
-    /// UnexpectedEos (cause; assertion failed: `!buf.is_eos()`)
+    /// UnexpectedEos (cause; assertion failed: `left == right`; \
+    ///                assertion failed: `(left == right)` (left: `1`, right: `0`))
     /// HISTORY:
-    ///   [0] at src/bytes.rs:144
-    ///   [1] at src/fixnum.rs:199
-    ///   [2] at src/decode.rs:11 -- oops!
-    ///   [3] at src/decode.rs:17\n");
+    ///   [0] at src/bytes.rs:140
+    ///   [1] at src/fixnum.rs:196
+    ///   [2] at src/decode.rs:12 -- oops!
+    ///   [3] at src/io.rs:44
+    ///   [4] at src/decode.rs:16\n");
     /// # }
     /// ```
     fn map_err<F, E>(self, f: F) -> MapErr<Self, F, E>
@@ -233,15 +198,14 @@ pub trait DecodeExt: Decode + Sized {
     /// Decodes a length-prefixed string:
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt};
+    /// use bytecodec::{Decode, DecodeExt};
     /// use bytecodec::bytes::Utf8Decoder;
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = U8Decoder::new().and_then(|len| Utf8Decoder::new().length(len as u64));
-    /// let mut input = DecodeBuf::new(b"\x03foobar");
-    ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some("foo".to_owned()));
+    /// let item = decoder.decode_exact(b"\x03foobar".as_ref()).unwrap();
+    /// assert_eq!(item, "foo");
     /// ```
     fn and_then<D, F>(self, f: F) -> AndThen<Self, D, F>
     where
@@ -258,21 +222,22 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt, StartDecoderChain};
+    /// use bytecodec::{Decode, DecodeExt, StartDecoderChain};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = StartDecoderChain
     ///     .chain(U8Decoder::new())
     ///     .chain(U8Decoder::new())
     ///     .chain(U8Decoder::new());
     ///
-    /// let mut input = DecodeBuf::new(b"foobar");
+    /// let mut input = &b"foobar"[..];
     ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some((b'f', b'o', b'o')));
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, (b'f', b'o', b'o'));
     ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some((b'b', b'a', b'r')));
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, (b'b', b'a', b'r'));
     /// ```
     fn chain<D: Decode>(self, other: D) -> DecoderChain<Self, D, Self::Item> {
         DecoderChain::new(self, other)
@@ -283,14 +248,13 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt};
+    /// use bytecodec::{Decode, DecodeExt};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = U8Decoder::new().collect::<Vec<_>>();
-    /// let mut input = DecodeBuf::with_remaining_bytes(b"foo", 0);
-    ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some(vec![b'f', b'o', b'o']));
+    /// let item = decoder.decode_exact(b"foo".as_ref()).unwrap();
+    /// assert_eq!(item, vec![b'f', b'o', b'o']);
     /// ```
     fn collect<T>(self) -> Collect<Self, T>
     where
@@ -304,19 +268,20 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt, ErrorKind};
+    /// use bytecodec::{Decode, DecodeExt, ErrorKind};
     /// use bytecodec::bytes::Utf8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = Utf8Decoder::new().length(3);
-    /// let mut input = DecodeBuf::with_remaining_bytes(b"foobarba", 0);
+    /// let mut input = &b"foobarba"[..];
     ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some("foo".to_owned()));
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, "foo");
     ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some("bar".to_owned()));
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, "bar");
     ///
-    /// let error = decoder.decode(&mut input).err().unwrap();
+    /// let error = decoder.decode_exact(&mut input).err().unwrap();
     /// assert_eq!(*error.kind(), ErrorKind::UnexpectedEos);
     /// ```
     fn length(self, expected_bytes: u64) -> Length<Self> {
@@ -328,14 +293,13 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt};
+    /// use bytecodec::{Decode, DecodeExt};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = U8Decoder::new().take(2).collect::<Vec<_>>();
-    /// let mut input = DecodeBuf::new(b"foo");
-    ///
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some(vec![b'f', b'o']));
+    /// let item = decoder.decode_exact(b"foo".as_ref()).unwrap();
+    /// assert_eq!(item, vec![b'f', b'o']);
     /// ```
     fn take(self, n: usize) -> Take<Self> {
         Take::new(self, n)
@@ -346,18 +310,19 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt};
+    /// use bytecodec::{Decode, DecodeExt};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
-    /// let mut input = DecodeBuf::new(b"foo");
+    /// let mut input = &b"foo"[..];
     ///
     /// let mut decoder = U8Decoder::new().omit(true);
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some(None));
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, None);
     ///
     /// let mut decoder = U8Decoder::new().omit(false);
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some(Some(b'f')));
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, Some(b'f'));
     /// ```
     fn omit(self, do_omit: bool) -> Omit<Self> {
         Omit::new(self, do_omit)
@@ -369,15 +334,16 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt};
+    /// use bytecodec::{Decode, DecodeExt};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
-    /// let mut input = DecodeBuf::with_remaining_bytes(b"foo", 0);
+    /// let mut input = &b"foo"[..];
     ///
     /// let mut decoder = U8Decoder::new().skip_remaining();
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some(b'f'));
-    /// assert!(input.is_empty() && input.is_eos());
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, b'f');
+    /// assert!(input.is_empty());
     /// ```
     fn skip_remaining(self) -> SkipRemaining<Self> {
         SkipRemaining::new(self)
@@ -388,17 +354,16 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt, ErrorKind};
+    /// use bytecodec::{Decode, DecodeExt, ErrorKind};
     /// use bytecodec::bytes::Utf8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = Utf8Decoder::new().max_bytes(3);
     ///
-    /// let mut input = DecodeBuf::with_remaining_bytes(b"foo", 0);
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some("foo".to_owned())); // OK
+    /// let item = decoder.decode_exact(b"foo".as_ref()).unwrap();
+    /// assert_eq!(item, "foo"); // OK
     ///
-    /// let mut input = DecodeBuf::with_remaining_bytes(b"hello", 0);
-    /// let error = decoder.decode(&mut input).err();
+    /// let error = decoder.decode_exact(b"hello".as_ref()).err();
     /// assert_eq!(error.map(|e| *e.kind()), Some(ErrorKind::InvalidInput)); // Error
     /// ```
     fn max_bytes(self, bytes: u64) -> MaxBytes<Self> {
@@ -410,16 +375,17 @@ pub trait DecodeExt: Decode + Sized {
     /// # Examples
     ///
     /// ```
-    /// use bytecodec::{Decode, DecodeBuf, DecodeExt, ErrorKind};
+    /// use bytecodec::{Decode, DecodeExt, ErrorKind};
     /// use bytecodec::fixnum::U8Decoder;
+    /// use bytecodec::io::IoDecodeExt;
     ///
     /// let mut decoder = U8Decoder::new().assert(|&b| b == 3);
     ///
-    /// let mut input = DecodeBuf::new(&[3, 4][..]);
-    /// let item = decoder.decode(&mut input).unwrap();
-    /// assert_eq!(item, Some(3));
+    /// let mut input = &[3, 4][..];
+    /// let item = decoder.decode_exact(&mut input).unwrap();
+    /// assert_eq!(item, 3);
     ///
-    /// let error = decoder.decode(&mut input).err();
+    /// let error = decoder.decode_exact(&mut input).err();
     /// assert_eq!(error.map(|e| *e.kind()), Some(ErrorKind::InvalidInput));
     /// ```
     fn assert<F>(self, f: F) -> Assert<Self, F>
@@ -428,5 +394,44 @@ pub trait DecodeExt: Decode + Sized {
     {
         Assert::new(self, f)
     }
+
+    /// Creates a decoder that makes it possible to slice the input byte sequence in arbitrary units.
+    ///
+    /// Slicing an input byte sequence makes it easier to demultiplex multiple sequences from it.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use bytecodec::{Decode, DecodeExt, Eos};
+    /// use bytecodec::bytes::Utf8Decoder;
+    ///
+    /// let mut decoder0 = Utf8Decoder::new().length(3).slice();
+    /// let mut decoder1 = Utf8Decoder::new().length(3).slice();
+    ///
+    /// let eos = Eos::new(true);
+    /// let input = b"fboaor";
+    /// let mut offset = 0;
+    ///
+    /// let mut last_item0 = None;
+    /// let mut last_item1 = None;
+    /// for _ in 0..3 {
+    ///     decoder0.set_consumable_bytes(1);
+    ///     let (size, item) = decoder0.decode(&input[offset..], eos).unwrap();
+    ///     offset += size;
+    ///     last_item0 = item;
+    ///
+    ///     decoder1.set_consumable_bytes(1);
+    ///     let (size, item) = decoder1.decode(&input[offset..], eos).unwrap();
+    ///     offset += size;
+    ///     last_item1 = item;
+    /// }
+    ///
+    /// assert_eq!(offset, input.len());
+    /// assert_eq!(last_item0, Some("foo".to_owned()));
+    /// assert_eq!(last_item1, Some("bar".to_owned()));
+    /// ```
+    fn slice(self) -> Slice<Self> {
+        Slice::new(self)
+    }
 }
 impl<T: Decode> DecodeExt for T {}
